// awy_dat.cxx
// load FG awy.dat.gz
/* 
  sample
start  latst     lonst       end    latend    lonend      t base top name
00MKK  22.528056 -156.170961 BITTA  23.528031 -155.478836 1 012 460 R464
00MKK  22.528056 -156.170961 CKH99  22.316668 -156.341660 1 012 460 R464
00UPP  20.566668 -154.125000 BYROW  20.369456 -155.058992 1 140 460 R578
00UPP  20.566668 -154.125000 FITES  20.794556 -153.000633 1 210 460 R578
  // Read each line from the database
  identStart = token;
  in >> latStart >> lonStart >> identEnd >> latEnd >> lonEnd >>
  type >> base >> top >> name;

==== */

//#include "aptdat.h"
#include <string>
#include <iostream>
#include "gshhs_win.h"
#include "zlib.h"

#ifndef  NO_ADD_FG_DATA

using namespace std;

LE sAwy_List = { &sAwy_List, &sAwy_List };
static int awyindex = 0;

PLE Get_Awy_List( void )
{
   return &sAwy_List;
}

void Del_Awy_List( void )
{
   PLE ph = Get_Awy_List();
   KillLList(ph);
}

#define MSTRCPY(a,b,c)  if(strlen(b) < c )\
   strcpy(a,b);\
   else {\
      strncpy(a,b,c);\
      a[c] = 0;\
}

PFGAWY Add_Awy_List( PLE ph, const char * idstart, double latst, double lonst,
                    const char * idend, double latend, double lonend,
                    int type, int base, int top, const char * name )
{
   PFGAWY paw = (PFGAWY)MALLOC( sizeof(FGAWY) );
   CHKMEM(paw, sizeof(FGAWY));
   awyindex++;
   paw->index = awyindex;
   MSTRCPY(paw->id_start, idstart, MX_FIX_NAME);
   paw->latst = latst;
   paw->lonst = lonst;
   MSTRCPY( paw->id_end, idend, MX_FIX_NAME);
   paw->latend = latend;
   paw->lonend = lonend;
   paw->type   = type;
   paw->base   = base;
   paw->top    = top;
   MSTRCPY( paw->name, name, MX_FIX_NAME );
   InsertTailList(ph, (PLE)paw);
   return paw;
}

PFGAWY Add_Awy_Item( const char * idstart, double latst, double lonst,
                    const char * idend, double latend, double lonend,
                    int type, int base, int top, const char * name )
{
   PLE ph = Get_Awy_List();
   return Add_Awy_List( ph, idstart, latst, lonst,
                    idend, latend, lonend,
                    type, base, top, name );
}


static char * tokens[FG_MAX_TOKENS];
static int tok_count;
static char buf[FG_MX_BUFF+8];
int Load_Awy_File( char * pfile )
{
   int   added_count = 0;
   unsigned int line_num = 0;
   gzFile gzf = gzopen(pfile, "rb");
   if( gzf ) {
      char * pb = buf;
      int    line_num = 0;
      char * cp;
      string name = "";
      string line = "";
      // skip first two lines
      cp = gzgets( gzf, pb, FG_MX_BUFF ); // int rd = gzread( gzf, buf, 2048 );
      cp = gzgets( gzf, pb, FG_MX_BUFF ); // int rd = gzread( gzf, buf, 2048 );
      cp = gzgets( gzf, pb, FG_MX_BUFF ); // int rd = gzread( gzf, buf, 2048 );
      awyindex = 0;  // start INDEXER
      do {
         size_t len = strlen(buf);
         while(len--) {
            if( buf[len] > ' ' )
               break;
            buf[len] = 0;
         }
         cp = buf;
         while( *cp && (*cp <= ' '))
            cp++;
         len = strlen(cp);
         line_num++;
         tok_count = 0;
         tokens[tok_count++] = cp;
         if( len > 0 ) {
            line = pb;
            while( *cp ) {
               while( *cp > ' ' ) cp++;
               if( *cp ) {
                  *cp = 0;
                  cp++;
                  while( *cp && (*cp <= ' ') ) cp++;
                  tokens[tok_count++] = cp;
                  if(tok_count >= FG_MAX_TOKENS )
                     break;
               }
            }
            if( strcmp( tokens[0], "99" ) == 0 ) {
               if( iVerbal >= 2 ) {
                  sprtf( "End of list reached ...\n" );
               }
               break;
            }
            if( tok_count > 9 ) {
               //  identStart = token;
               //  in >> latStart >> lonStart >> identEnd >> latEnd >> lonEnd >>
               //  type >> base >> top >> name;
               string idstart = tokens[0];
               double latst = atof( tokens[1] );
               double lonst = atof( tokens[2] );
               string idend = tokens[3];
               double latend = atof( tokens[4] );
               double lonend = atof( tokens[5] );
               int type = atoi( tokens[6] );
               int base = atoi( tokens[7] );
               int top  = atoi( tokens[8] );
               name = tokens[9];
               Add_Awy_Item( idstart.c_str(), latst, lonst, idend.c_str(),
                  latend, lonend, type, base, top, name.c_str() );
               added_count++;
            }
         }  // got length
      } while(( cp = gzgets( gzf, pb, FG_MX_BUFF ) ) != NULL ); 
      gzclose(gzf);
      sprtf( "Added %d AIRWAYS to list ...\n", added_count );
   } else {
      sprtf( "WARNING: Failed to open %s file ...\n", pfile );
   }
   return added_count;
}

int Get_Awy_in_Bounds( double n, double s, double e, double w, PLE ph )
{
   int   found = 0;
   PLE   pha = Get_Awy_List();
   PLE   pn;
   double latst, lonst;
   double latend, lonend;
   PFGAWY pfa, pfa2;

   Traverse_List( pha, pn )
   {
      pfa = (PFGAWY)pn;
      latst = pfa->latst;
      lonst = pfa->lonst;
      latend = pfa->latend;
      lonend = pfa->lonend;
      if((( latst <= n )&&
         ( latst >= s )&&
         ( lonst <= e )&&
         ( lonst >= w ))||
         (( latend <= n )&&
         ( latend >= s )&&
         ( lonend <= e )&&
         ( lonend >= w )))
      {
         found++;
         pfa2 = (PFGAWY)MALLOC(sizeof(FGAWY));
         CHKMEM(pfa2, sizeof(FGAWY));
         memcpy(pfa2, pfa, sizeof(FGAWY));
         InsertTailList(ph, (PLE)pfa2);
      }
   }
   return found;
}

#endif // #ifndef  NO_ADD_FG_DATA

// eof - awy_dat.cxx
